@@ -16,6 +16,20 @@ bt='`'               # backtick
 start_code_block='```'
 end_code_block='```'
 
+# We put local binaries such as opam in .bin and extend PATH
+BIN=$(pwd)/.bin
+mkdir "$BIN"
+wget https://github.com/ocaml/opam/releases/download/2.1.3/opam-2.1.3-x86_64-linux -O "$BIN"/opam
+chmod +x "$BIN"/opam
+
+export PATH="$BIN":$PATH
+
+echo "Global env info:"
+echo "----------------"
+echo "pwd: $PWD"
+echo "path: $PATH"
+echo "opam version: `opam --version`"
+
 number_of_processors=$(cat /proc/cpuinfo | grep '^processor *' | wc -l)
 
 program_name="$0"
@@ -35,8 +49,6 @@ check_variable () {
   fi
 }
 
-echo $PWD
-
 #check_variable "BUILD_ID"
 #check_variable "BUILD_URL"
 #check_variable "JOB_NAME"
@@ -55,7 +67,8 @@ echo $PWD
 : "${old_coq_opam_archive_git_branch:=master}"
 : "${num_of_iterations:=1}"
 : "${timeout:=3h}"
-: "${coq_opam_packages:=coq-bignums coq-hott coq-performance-tests-lite coq-engine-bench-lite coq-mathcomp-ssreflect coq-mathcomp-fingroup coq-mathcomp-algebra coq-mathcomp-solvable coq-mathcomp-field coq-mathcomp-character coq-mathcomp-odd-order coq-math-classes coq-corn coq-flocq3 coq-compcert coq-metacoq-template coq-metacoq-pcuic coq-metacoq-safechecker coq-metacoq-erasure coq-metacoq-translations coq-geocoq coq-color coq-coqprime coq-coqutil coq-bedrock2 coq-rewriter coq-fiat-core coq-fiat-parsers coq-fiat-crypto-with-bedrock coq-unimath coq-coquelicot coq-iris-examples coq-verdi coq-verdi-raft coq-fourcolor coq-rewriter-perf-SuperFast coq-perennial coq-vst coq-category-theory}"
+: "${coq_opam_packages:=coq-bignums coq-hott coq-performance-tests-lite coq-engine-bench-lite coq-mathcomp-ssreflect coq-mathcomp-fingroup coq-mathcomp-algebra coq-mathcomp-solvable coq-mathcomp-field coq-mathcomp-character coq-mathcomp-odd-order coq-math-classes coq-corn coq-compcert coq-metacoq-template coq-metacoq-pcuic coq-metacoq-safechecker coq-metacoq-erasure coq-metacoq-translations coq-geocoq coq-color coq-coqprime coq-coqutil coq-bedrock2 coq-rewriter coq-fiat-core coq-fiat-parsers coq-fiat-crypto-with-bedrock coq-unimath coq-coquelicot coq-iris-examples coq-verdi coq-verdi-raft coq-fourcolor coq-rewriter-perf-SuperFast coq-perennial coq-vst coq-category-theory}"
+: "${coq_native:=}"
 
 new_coq_commit=$(git rev-parse HEAD^2)
 old_coq_commit=$(git merge-base HEAD^1 $new_coq_commit)
@@ -93,6 +106,7 @@ echo "DEBUG: num_of_iterations = $num_of_iterations"
 echo "DEBUG: coq_opam_packages = $coq_opam_packages"
 echo "DEBUG: coq_pr_number = $coq_pr_number"
 echo "DEBUG: coq_pr_comment_id = $coq_pr_comment_id"
+echo "DEBUG: coq_native = $coq_native"
 
 # --------------------------------------------------------------------------------
 
@@ -257,7 +271,7 @@ $ending
     fi
     curl -sSX PATCH https://coq.zulipchat.com/api/v1/messages/"$zulip_post" \
          -u "$ZULIP_BENCH_BOT" \
-         --data-urlencode content="$msg" >/dev/null
+         --data-urlencode content="$msg" >/dev/null || echo "Failed to edit zulip post" >&2
 }
 zulip_autofail() {
     code=$?
@@ -266,6 +280,9 @@ zulip_autofail() {
 }
 if [[ $zulip_post ]]; then trap zulip_autofail ERR; fi
 
+# see https://github.com/coq/coq/pull/15807
+ulimit -S -s $((2 * $(ulimit -s)))
+
 # Clone the indicated git-repository.
 
 coq_dir="$working_dir/coq"
@@ -321,7 +338,7 @@ create_opam() {
 
     # For some reason opam guesses an incorrect upper bound on the
     # number of jobs available on Travis, so we set it here manually:
-    opam config set-global jobs $number_of_processors
+    opam var --global jobs=$number_of_processors >/dev/null
     if [ ! -z "$BENCH_DEBUG" ]; then opam config list; fi
 
     opam repo add -q --this-switch coq-extra-dev "$OPAM_COQ_DIR/extra-dev"
@@ -343,7 +360,9 @@ create_opam() {
 
     echo "$1_coq_commit_long = $COQ_HASH_LONG"
 
-    for package in coq-core coq-stdlib coq; do
+    if [ ! -z "$coq_native" ]; then opam install coq-native; fi
+
+    for package in coq-core coq-stdlib coqide-server coq; do
         export COQ_OPAM_PACKAGE=$package
         export COQ_ITERATION=1
         _RES=0
@@ -430,13 +449,13 @@ for coq_opam_package in $sorted_coq_opam_packages; do
 
         # OPAM 2.0 likes to ignore the -j when it feels like :S so we
         # workaround that here.
-        opam config set-global jobs $number_of_processors
+        opam var --global jobs=$number_of_processors >/dev/null
 
         opam install $coq_opam_package -v -b -j$number_of_processors --deps-only -y \
              3>$log_dir/$coq_opam_package.$RUNNER.opam_install.deps_only.stdout.log 1>&3 \
              4>$log_dir/$coq_opam_package.$RUNNER.opam_install.deps_only.stderr.log 2>&4 || continue 2
 
-        opam config set-global jobs 1
+        opam var --global jobs=1 >/dev/null
 
         if [ ! -z "$BENCH_DEBUG" ]; then ls -l $working_dir; fi
 
