@@ -19,7 +19,6 @@ open Libnames
 open Namegen
 open Impargs
 open CAst
-open Notation_term
 open Notation
 open Constrexpr
 open Constrexpr_ops
@@ -29,6 +28,7 @@ open Glob_ops
 open Pattern
 open Detyping
 open Structures
+open Notationextern
 
 module NamedDecl = Context.Named.Declaration
 (*i*)
@@ -90,123 +90,6 @@ let is_reserved_type na t =
 
 (**********************************************************************)
 (* Turning notations and scopes on and off for printing *)
-module IRuleSet = Set.Make(struct
-    type t = interp_rule
-    let compare x y = match x, y with
-      | AbbrevRule kn, AbbrevRule kn' -> KerName.compare kn kn'
-      | NotationRule (scope,ntn), NotationRule (scope',ntn') -> compare x y
-      | AbbrevRule _, NotationRule _ -> -1
-      | NotationRule _, AbbrevRule _ -> 1
-  end)
-
-let inactive_notations_table =
-  Summary.ref ~name:"inactive_notations_table" (IRuleSet.empty)
-let inactive_scopes_table    =
-  Summary.ref ~name:"inactive_scopes_table" CString.Set.empty
-
-let show_scope inscope =
-  match inscope with
-  | LastLonelyNotation -> str ""
-  | NotationInScope sc -> spc () ++ str "in scope" ++ spc () ++ str sc
-
-let _show_inactive_notations () =
-  begin
-    if CString.Set.is_empty !inactive_scopes_table
-    then
-      Feedback.msg_notice (str "No inactive notation scopes.")
-    else
-      let _ = Feedback.msg_notice (str "Inactive notation scopes:") in
-      CString.Set.iter (fun sc -> Feedback.msg_notice (str "  " ++ str sc))
-        !inactive_scopes_table
-  end;
-  if IRuleSet.is_empty !inactive_notations_table
-  then
-    Feedback.msg_notice (str "No individual inactive notations.")
-  else
-    let _ = Feedback.msg_notice (str "Inactive notations:") in
-    IRuleSet.iter
-      (function
-       | NotationRule (inscope, ntn) ->
-         Feedback.msg_notice (pr_notation ntn ++ show_scope inscope)
-       | AbbrevRule kn -> Feedback.msg_notice (str (string_of_qualid (Nametab.shortest_qualid_of_abbreviation Id.Set.empty kn))))
-      !inactive_notations_table
-
-let deactivate_notation nr =
-  match nr with
-  | AbbrevRule kn ->
-     (* shouldn't we check whether it is well defined? *)
-     inactive_notations_table := IRuleSet.add nr !inactive_notations_table
-  | NotationRule (inscope, ntn) ->
-     let scopt = match inscope with NotationInScope sc -> Some sc | LastLonelyNotation -> None in
-     match availability_of_notation (inscope, ntn) (scopt, []) with
-     | None -> user_err
-                        (pr_notation ntn ++ spc () ++ str "does not exist"
-                         ++ (match inscope with
-                             | LastLonelyNotation -> spc () ++ str "in the empty scope."
-                             | NotationInScope _ -> show_scope inscope ++ str "."))
-     | Some _ ->
-        if IRuleSet.mem nr !inactive_notations_table then
-          Feedback.msg_warning
-            (str "Notation" ++ spc () ++ pr_notation ntn ++ spc ()
-             ++ str "is already inactive" ++ show_scope inscope ++ str ".")
-        else inactive_notations_table := IRuleSet.add nr !inactive_notations_table
-
-let reactivate_notation nr =
-  try
-    inactive_notations_table :=
-      IRuleSet.remove nr !inactive_notations_table
-  with Not_found ->
-    match nr with
-    | NotationRule (inscope, ntn) ->
-       Feedback.msg_warning (str "Notation" ++ spc () ++ pr_notation ntn ++ spc ()
-                             ++ str "is already active" ++ show_scope inscope ++
-  str ".")
-    | AbbrevRule kn ->
-       let s = string_of_qualid (Nametab.shortest_qualid_of_abbreviation Id.Set.empty kn) in
-       Feedback.msg_warning
-         (str "Notation" ++ spc () ++ str s
-          ++ spc () ++ str "is already active.")
-
-
-let deactivate_scope sc =
-  ignore (find_scope sc); (* ensures that the scope exists *)
-  if CString.Set.mem sc !inactive_scopes_table
-  then
-    Feedback.msg_warning (str "Notation Scope" ++ spc () ++ str sc ++ spc ()
-                          ++ str "is already inactive.")
-  else
-    inactive_scopes_table := CString.Set.add sc !inactive_scopes_table
-
-let reactivate_scope sc =
-  try
-    inactive_scopes_table := CString.Set.remove sc !inactive_scopes_table
-  with Not_found ->
-    Feedback.msg_warning (str "Notation Scope" ++ spc () ++ str sc ++ spc ()
-                          ++ str "is already active.")
-
-let is_inactive_rule nr =
-  IRuleSet.mem nr !inactive_notations_table ||
-  match nr with
-    | NotationRule (NotationInScope sc, ntn) -> CString.Set.mem sc !inactive_scopes_table
-    | NotationRule (LastLonelyNotation, ntn) -> false
-    | AbbrevRule _ -> false
-
-(* args: notation, scope, activate/deactivate *)
-let toggle_scope_printing ~scope ~activate =
-  if activate then
-    reactivate_scope scope
-  else
-    deactivate_scope scope
-
-let toggle_notation_printing ?scope ~notation ~activate () =
-  let inscope = match scope with
-  | None -> LastLonelyNotation
-  | Some sc -> NotationInScope sc in
-  if activate then
-    reactivate_notation (NotationRule (inscope, notation))
-  else
-    deactivate_notation (NotationRule (inscope, notation))
-
 (* This governs printing of projections using the dot notation symbols *)
 let print_projections = ref false
 
@@ -216,10 +99,6 @@ let with_universes f = Flags.with_option print_universes f
 let with_meta_as_hole f = Flags.with_option print_meta_as_hole f
 let without_symbols f = Flags.with_option print_no_symbol f
 
-let without_specific_symbols l =
-  Flags.with_modified_ref inactive_notations_table
-    (fun tbl -> IRuleSet.(union (of_list l) tbl))
-
 (**********************************************************************)
 (* Control printing of records *)
 
@@ -338,7 +217,7 @@ let rec fill_arg_scopes args subscopes (entry,(_,scopes) as all) =
   | a :: args, scopt :: subscopes ->
     (a, (entry, (scopt, scopes))) :: fill_arg_scopes args subscopes all
   | a :: args, [] ->
-    (a, (entry, (None, scopes))) :: fill_arg_scopes args [] all
+    (a, (entry, ([], scopes))) :: fill_arg_scopes args [] all
 
 (**********************************************************************)
 (* mapping patterns to cases_pattern_expr                                *)
@@ -571,7 +450,7 @@ and extern_notation_pattern allscopes vars t = function
   | [] -> raise No_match
   | (keyrule,pat,n as _rule)::rules ->
     try
-      if is_inactive_rule keyrule || is_printing_inactive_rule keyrule pat then raise No_match;
+      if is_printing_inactive_rule keyrule pat then raise No_match;
       let loc = t.loc in
       match DAst.get t with
         | PatCstr (cstr,args,na) ->
@@ -588,7 +467,7 @@ let rec extern_notation_ind_pattern allscopes vars ind args = function
   | [] -> raise No_match
   | (keyrule,pat,n as _rule)::rules ->
     try
-      if is_inactive_rule keyrule || is_printing_inactive_rule keyrule pat then raise No_match;
+      if is_printing_inactive_rule keyrule pat then raise No_match;
       apply_notation_to_pattern (GlobRef.IndRef ind)
         (match_notation_constr_ind_pattern ind args pat) allscopes vars keyrule
     with
@@ -614,7 +493,7 @@ let extern_ind_pattern_in_scope (custom,scopes as allscopes) vars ind args =
            |None           -> CAst.make @@ CPatCstr (c, Some args, [])
 
 let extern_cases_pattern vars p =
-  extern_cases_pattern_in_scope (InConstrEntrySomeLevel,(None,[])) vars p
+  extern_cases_pattern_in_scope (InConstrEntrySomeLevel,([],[])) vars p
 
 (**********************************************************************)
 (* Externalising applications *)
@@ -1164,9 +1043,9 @@ let rec extern inctx ?impargs scopes vars r =
   in insert_entry_coercion coercion (CAst.make ?loc c)
 
 and extern_typ ?impargs (subentry,(_,scopes)) =
-  extern true ?impargs (subentry,(Notation.current_type_scope_name (),scopes))
+  extern true ?impargs (subentry,(Notation.current_type_scope_names (),scopes))
 
-and sub_extern inctx (subentry,(_,scopes)) = extern inctx (subentry,(None,scopes))
+and sub_extern inctx (subentry,(_,scopes)) = extern inctx (subentry,([],scopes))
 
 and factorize_prod ?impargs scopes vars na bk t c =
   let implicit_type = is_reserved_type na t in
@@ -1298,7 +1177,7 @@ and extern_notation inctx (custom,scopes as allscopes) vars t rules =
   | (keyrule,pat,n as _rule)::rules ->
       let loc = Glob_ops.loc_of_glob_constr t in
       try
-        if is_inactive_rule keyrule || is_printing_inactive_rule keyrule pat then raise No_match;
+        if is_printing_inactive_rule keyrule pat then raise No_match;
         let f,args =
           match DAst.get t with
           | GApp (f,args) -> f,args
@@ -1414,10 +1293,10 @@ and extern_applied_proj inctx scopes vars (cst,us) params c extraargs =
   extern_projection inctx (f,us) nparams args imps
 
 let extern_glob_constr vars c =
-  extern false (InConstrEntrySomeLevel,(None,[])) vars c
+  extern false (InConstrEntrySomeLevel,([],[])) vars c
 
 let extern_glob_type ?impargs vars c =
-  extern_typ ?impargs (InConstrEntrySomeLevel,(None,[])) vars c
+  extern_typ ?impargs (InConstrEntrySomeLevel,([],[])) vars c
 
 (******************************************************************)
 (* Main translation function from constr -> constr_expr *)
@@ -1425,6 +1304,7 @@ let extern_glob_type ?impargs vars c =
 let extern_constr ?lax ?(inctx=false) ?scope env sigma t =
   let r = Detyping.detype Detyping.Later ?lax false Id.Set.empty env sigma t in
   let vars = extern_env env sigma in
+  let scope = Option.cata (fun x -> [x]) [] scope in
   extern inctx (InConstrEntrySomeLevel,(scope,[])) vars r
 
 let extern_constr_in_scope ?lax ?inctx scope env sigma t =
@@ -1450,6 +1330,7 @@ let extern_closed_glob ?lax ?(goal_concl_style=false) ?(inctx=false) ?scope env
     Detyping.detype_closed_glob ?lax goal_concl_style avoid env sigma t
   in
   let vars = extern_env env sigma in
+  let scope = Option.cata (fun x -> [x]) [] scope in
   extern inctx (InConstrEntrySomeLevel,(scope,[])) vars r
 
 (******************************************************************)
@@ -1467,8 +1348,10 @@ let rec glob_of_pat avoid env sigma pat = DAst.make @@ match pat with
   | PRef ref -> GRef (ref,None)
   | PVar id  -> GVar id
   | PEvar (evk,l) ->
-      let test decl = function PVar id' -> Id.equal (NamedDecl.get_id decl) id' | _ -> false in
-      let l = Evd.evar_instance_array test (Evd.find sigma evk) l in
+      let filter (id, pat) = match pat with PVar id' -> Id.equal id id' | _ -> true in
+      let hyps = Evd.evar_filtered_context (Evd.find sigma evk) in
+      let map decl pat = NamedDecl.get_id decl, pat in
+      let l = List.filter filter @@ List.map2 map hyps l in
       let id = match Evd.evar_ident evk sigma with
       | None -> Id.of_string "__"
       | Some id -> id
@@ -1589,7 +1472,7 @@ and glob_of_pat_under_context avoid env sigma (nas, pat) =
   (Array.rev_of_list nas, pat)
 
 let extern_constr_pattern env sigma pat =
-  extern true (InConstrEntrySomeLevel,(None,[]))
+  extern true (InConstrEntrySomeLevel,([],[]))
     (* XXX no vars? *)
     (Id.Set.empty, Evd.universe_binders sigma)
     (glob_of_pat Id.Set.empty env sigma pat)
@@ -1598,4 +1481,4 @@ let extern_rel_context where env sigma sign =
   let a = detype_rel_context Detyping.Later where Id.Set.empty (names_of_rel_context env,env) sigma sign in
   let vars = extern_env env sigma in
   let a = List.map (extended_glob_local_binder_of_decl) a in
-  pi3 (extern_local_binder (InConstrEntrySomeLevel,(None,[])) vars a)
+  pi3 (extern_local_binder (InConstrEntrySomeLevel,([],[])) vars a)
