@@ -95,7 +95,7 @@ let evaluable_reference_eq sigma r1 r2 = match r1, r2 with
 | EvalVar id1, EvalVar id2 -> Id.equal id1 id2
 | EvalRel i1, EvalRel i2 -> Int.equal i1 i2
 | EvalEvar (e1, ctx1), EvalEvar (e2, ctx2) ->
-  Evar.equal e1 e2 && List.equal (EConstr.eq_constr sigma) ctx1 ctx2
+  EConstr.eq_constr sigma (mkEvar (e1, ctx1)) (mkEvar (e2, ctx2))
 | _ -> false
 
 let mkEvalRef ref u =
@@ -392,7 +392,7 @@ let substl_with_function subst sigma constr =
         let (sigma, evk) = Evarutil.new_pure_evar empty_named_context_val sigma dummy in
         evd := sigma;
         minargs := Evar.Map.add evk (min, fx, ref) !minargs;
-        mkEvar (evk, [])
+        mkEvar (evk, SList.empty)
       | (fx, None) -> Vars.lift k fx
     else mkRel (i - Array.length v)
   | _ ->
@@ -439,7 +439,7 @@ let substl_checking_arity env subst sigma c =
   (* we propagate the constraints: solved problems are substituted;
      the other ones are replaced by the function symbol *)
   let rec nf_fix k c = match EConstr.kind sigma c with
-  | Evar (i, []) ->
+  | Evar (i, SList.Nil) ->
     (* FIXME: find a less hackish way of doing this *)
     begin match Evar.Map.find i minargs with
     | (_, fx, ref) ->
@@ -1029,10 +1029,14 @@ let whd_simpl_orelse_delta_but_fix env sigma c =
       | _ -> redrec (c, stack))
     | None -> s'
   in
-  let simpfun = clos_norm_flags betaiota env sigma in
-  simpfun (applist (redrec c))
+  applist (redrec c)
 
-let hnf_constr env sigma c = whd_simpl_orelse_delta_but_fix env sigma (c, [])
+let hnf_constr0 env sigma c =
+  whd_simpl_orelse_delta_but_fix env sigma (c, [])
+
+let hnf_constr env sigma c =
+  let c = whd_simpl_orelse_delta_but_fix env sigma (c, []) in
+  clos_norm_flags betaiota env sigma c
 
 (* The "simpl" reduction tactic *)
 
@@ -1255,27 +1259,56 @@ let check_privacy env ind =
 
 let reduce_to_ind_gen allow_product env sigma t =
   let rec elimrec env t l =
-    let t = hnf_constr env sigma t in
+    let t = hnf_constr0 env sigma t in
     match EConstr.kind sigma (fst (decompose_app_vect sigma t)) with
-      | Ind (ind, _ as indu) -> check_privacy env ind; (indu, it_mkProd_or_LetIn t l)
+      | Ind (ind, _ as indu) ->
+        let t = nf_betaiota env sigma t in
+        check_privacy env ind; (Some indu, it_mkProd_or_LetIn t l)
       | Prod (n,ty,t') ->
           let open Context.Rel.Declaration in
           if allow_product then
+            let ty = nf_betaiota env sigma ty in
             elimrec (push_rel (LocalAssum (n,ty)) env) t' ((LocalAssum (n,ty))::l)
           else
-            user_err Pp.(str"Not an inductive definition.")
+            None, it_mkProd_or_LetIn t l
       | _ ->
           (* Last chance: we allow to bypass the Opaque flag (as it
              was partially the case between V5.10 and V8.1 *)
           let t' = whd_all env sigma t in
           match EConstr.kind sigma (fst (decompose_app_vect sigma t')) with
-            | Ind (ind, _ as indu) -> check_privacy env ind; (indu, it_mkProd_or_LetIn t' l)
-            | _ -> user_err Pp.(str"Not an inductive product.")
+            | Ind (ind, _ as indu) -> check_privacy env ind; (Some indu, it_mkProd_or_LetIn t' l)
+            | _ -> None, it_mkProd_or_LetIn t l
+
   in
   elimrec env t []
 
-let reduce_to_quantified_ind env sigma c = reduce_to_ind_gen true env sigma c
-let reduce_to_atomic_ind env sigma c = reduce_to_ind_gen false env sigma c
+let reduce_to_quantified_ind env sigma c =
+  match reduce_to_ind_gen true env sigma c with
+  | None, _ -> user_err Pp.(str"Not an inductive definition.")
+  | Some i, t -> i, t
+let reduce_to_atomic_ind env sigma c =
+  match reduce_to_ind_gen false env sigma c with
+  | None, _ -> user_err Pp.(str"Not an inductive definition.")
+  | Some i, t -> i, t
+
+let eval_to_quantified_ind env sigma t =
+  let rec elimrec env t =
+    let t = hnf_constr0 env sigma t in
+    match EConstr.kind sigma (fst (decompose_app_vect sigma t)) with
+      | Ind (ind, _ as indu) ->
+        let () = check_privacy env ind in
+        indu
+      | Prod (n,ty,t') ->
+        elimrec (push_rel (Context.Rel.Declaration.LocalAssum (n,ty)) env) t'
+      | _ ->
+          (* Last chance: we allow to bypass the Opaque flag (as it
+             was partially the case between V5.10 and V8.1 *)
+          let t' = whd_all env sigma t in
+          match EConstr.kind sigma (fst (decompose_app_vect sigma t')) with
+            | Ind (ind, _ as indu) -> check_privacy env ind; indu
+            | _ -> user_err Pp.(str"Not an inductive product.")
+  in
+  elimrec env t
 
 let find_hnf_rectype env sigma t =
   let ind,t = reduce_to_atomic_ind env sigma t in
@@ -1323,24 +1356,26 @@ let error_cannot_recognize ref =
     Pp.(str "Cannot recognize a statement based on " ++
         Nametab.pr_global_env Id.Set.empty ref ++ str".")
 
-let reduce_to_ref_gen allow_product env sigma ref t =
-  if Globnames.isIndRef ref then
-    let ((mind,u),t) = reduce_to_ind_gen allow_product env sigma t in
-    begin match ref with
-    | GlobRef.IndRef mind' when Ind.CanOrd.equal mind mind' -> t
-    | _ -> error_cannot_recognize ref
-    end
-  else
-  (* lazily reduces to match the head of [t] with the expected [ref] *)
-  let rec elimrec env t l =
-    let c, _ = decompose_app_vect sigma t in
-    match EConstr.kind sigma c with
+let reduce_to_ref_gen allow_failure allow_product env sigma ref t =
+  match ref with
+  | GlobRef.IndRef mind' ->
+    let (i,t) = reduce_to_ind_gen allow_product env sigma t in
+    if allow_failure then t else
+      (match i with
+       | Some (mind,u) when Ind.CanOrd.equal mind mind' -> t
+       | _ -> error_cannot_recognize ref)
+  | _ -> (* lazily reduces to match the head of [t] with the expected [ref] *)
+    let rec elimrec env t l =
+      let c, _ = decompose_app_vect sigma t in
+      match EConstr.kind sigma c with
       | Prod (n,ty,t') ->
-          if allow_product then
-            let open Context.Rel.Declaration in
-            elimrec (push_rel (LocalAssum (n,ty)) env) t' ((LocalAssum (n,ty))::l)
-          else
-            error_cannot_recognize ref
+        if allow_product then
+          let open Context.Rel.Declaration in
+          elimrec (push_rel (LocalAssum (n,ty)) env) t' ((LocalAssum (n,ty))::l)
+        else if allow_failure then
+          it_mkProd_or_LetIn t l
+        else
+          error_cannot_recognize ref
       | _ ->
         if isRefX sigma ref c
         then it_mkProd_or_LetIn t l
@@ -1348,9 +1383,13 @@ let reduce_to_ref_gen allow_product env sigma ref t =
           try
             let t' = nf_betaiota env sigma (one_step_reduce env sigma t) in
             elimrec env t' l
-          with NotStepReducible -> error_cannot_recognize ref
-  in
-  elimrec env t []
+          with NotStepReducible ->
+            if allow_failure then
+              it_mkProd_or_LetIn t l
+            else
+              error_cannot_recognize ref
+    in
+    elimrec env t []
 
-let reduce_to_quantified_ref = reduce_to_ref_gen true
-let reduce_to_atomic_ref = reduce_to_ref_gen false
+let reduce_to_quantified_ref ?(allow_failure=false) = reduce_to_ref_gen allow_failure true
+let reduce_to_atomic_ref ?(allow_failure=false) = reduce_to_ref_gen allow_failure false
