@@ -151,10 +151,14 @@ type hint = {
   hint_arty : int; (* Number of goals generated by the intended tactic *)
 }
 
+type hint_pattern =
+| DefaultPattern
+| ConstrPattern of constr_pattern
+
 type 'a with_metadata =
   { pri     : int
   (** A number lower is higher priority *)
-  ; pat     : constr_pattern option
+  ; pat     : hint_pattern option
   (** A pattern for the concl of the Goal *)
   ; name    : hints_path_atom
   (** A potential name to refer to the hint *)
@@ -226,6 +230,12 @@ let pri_order_int (id1, {pri=pri1}) (id2, {pri=pri2}) =
 
 let pri_order t1 t2 = pri_order_int t1 t2 <= 0
 
+let get_default_pattern (h : hint hint_ast) = match h with
+| Give_exact h -> h.hint_type
+| Res_pf _ | ERes_pf _ | Res_pf_THEN_trivial_fail _ | Unfold_nth _ | Extern _ ->
+  (* Only exact hints may contain DefaultPattern *)
+  assert false
+
 (* Nov 98 -- Papageno *)
 (* Les Hints sont ré-organisés en plusieurs databases.
 
@@ -248,14 +258,15 @@ module Bounded_net :
 sig
   type t
   val empty : TransparentState.t option -> t
-  val add : t -> Pattern.constr_pattern -> stored_data -> t
+  val add : t -> hint_pattern -> stored_data -> t
   val lookup : Environ.env -> Evd.evar_map -> t -> EConstr.constr -> stored_data list
 end =
 struct
   module Data = struct type t = stored_data let compare = pri_order_int end
   module Bnet = Btermdn.Make(Data)
 
-  type diff = Pattern.constr_pattern * stored_data
+  type diff = hint_pattern * stored_data
+
   type data = Bnet of (TransparentState.t option * Bnet.t) | Diff of diff * data ref
   type t = data ref
 
@@ -263,17 +274,22 @@ struct
 
   let add net p v = ref (Diff ((p, v), net))
 
-  let rec force env net = match !net with
+  let rec force env sigma net = match !net with
   | Bnet dn -> dn
   | Diff ((p, v), rem) ->
-    let st, dn = force env rem in
-    let p = Bnet.pattern env st p in
+    let st, dn = force env sigma rem in
+    let p = match p with
+    | ConstrPattern p -> Bnet.pattern env st p
+    | DefaultPattern ->
+      let c = get_default_pattern (snd v).code.obj in
+      Bnet.constr_pattern env sigma st c
+    in
     let dn = Bnet.add dn p v in
     let () = net := (Bnet (st, dn)) in
     st, dn
 
   let lookup env sigma net p =
-    let st, dn = force env net in
+    let st, dn = force env sigma net in
     Bnet.lookup env sigma st dn p
 end
 
@@ -817,9 +833,8 @@ let make_exact_entry env sigma info ?(name=PathAny) (c, cty, ctx) =
         in
         let pri = match info.hint_priority with None -> 0 | Some p -> p in
         let pat = match info.hint_pattern with
-        | Some pat -> snd pat
-        | None ->
-          Patternops.pattern_of_constr env sigma cty
+        | Some pat -> ConstrPattern (snd pat)
+        | None -> DefaultPattern
         in
         let h = { rhint_term = c; rhint_type = cty; rhint_uctx = ctx; rhint_arty = 0 } in
         (Some hd,
@@ -828,9 +843,10 @@ let make_exact_entry env sigma info ?(name=PathAny) (c, cty, ctx) =
            code = with_uid (Give_exact h); })
 
 let make_apply_entry env sigma hnf info ?(name=PathAny) (c, cty, ctx) =
-  let cty = if hnf then hnf_constr env sigma cty else cty in
+  let cty = if hnf then hnf_constr0 env sigma cty else cty in
   match EConstr.kind sigma cty with
   | Prod _ ->
+    let cty = if hnf then Reductionops.nf_betaiota env sigma cty else cty in
     let sigma' = merge_context_set_opt sigma ctx in
     let ce = mk_clenv_from env sigma' (c,cty) in
     let c' = clenv_type (* ~reduce:false *) ce in
@@ -843,9 +859,8 @@ let make_apply_entry env sigma hnf info ?(name=PathAny) (c, cty, ctx) =
     let hyps = nb_hyp sigma' cty in
     let pri = match info.hint_priority with None -> hyps + nmiss | Some p -> p in
     let pat = match info.hint_pattern with
-    | Some p -> snd p
-    | None ->
-      Patternops.pattern_of_constr env ce.evd c'
+    | Some p -> ConstrPattern (snd p)
+    | None -> ConstrPattern (Patternops.pattern_of_constr env ce.evd c')
     in
     let h = { rhint_term = c; rhint_type = cty; rhint_uctx = ctx; rhint_arty = hyps; } in
     if Int.equal nmiss 0 then
@@ -929,7 +944,7 @@ let make_extern pri pat tacast =
   in
   (hdconstr,
    { pri = pri;
-     pat = pat;
+     pat = Option.map (fun p -> ConstrPattern p) pat;
      name = PathAny;
      db = None;
      secvars = Id.Pred.empty; (* Approximation *)
@@ -956,7 +971,7 @@ let make_trivial env sigma ?(name=PathAny) r =
   let h = { rhint_term = c; rhint_type = t; rhint_uctx = ctx; rhint_arty = 0 } in
   (Some hd,
    { pri=1;
-     pat = Some (Patternops.pattern_of_constr env ce.evd (clenv_type ce));
+     pat = Some (ConstrPattern (Patternops.pattern_of_constr env ce.evd (clenv_type ce)));
      name = name;
      db = None;
      secvars = secvars_of_constr env sigma c;
@@ -1146,7 +1161,13 @@ let subst_autohint (subst, obj) =
     let k' = Option.Smart.map subst_key k in
     let env = Global.env () in
     let sigma = Evd.from_env env in
-    let pat' = Option.Smart.map (subst_pattern env sigma subst) data.pat in
+    let subst_hint_pattern = function
+    | DefaultPattern -> DefaultPattern
+    | ConstrPattern p as p0 ->
+      let p' = subst_pattern env sigma subst p in
+      if p' == p then p0 else ConstrPattern p'
+    in
+    let pat' = Option.Smart.map subst_hint_pattern data.pat in
     let code' = match data.code.obj with
       | Res_pf h ->
         let h' = subst_aux h in
@@ -1164,7 +1185,7 @@ let subst_autohint (subst, obj) =
           let ref' = subst_evaluable_reference subst ref in
           if ref==ref' then data.code.obj else Unfold_nth ref'
       | Extern (pat, tac) ->
-          let pat' = Option.Smart.map (subst_pattern env sigma subst) data.pat in
+          let pat' = Option.Smart.map (subst_pattern env sigma subst) pat in
           let tac' = Genintern.generic_substitute subst tac in
           if pat==pat' && tac==tac' then data.code.obj else Extern (pat', tac')
     in
@@ -1278,11 +1299,14 @@ let make_hint ~locality name action =
 
 let warn_deprecated_hint_without_locality =
   CWarnings.create ~name:"deprecated-hint-without-locality" ~category:"deprecated"
-    (fun () -> strbrk "The default value for hint locality is currently \
-    \"local\" in a section and \"global\" otherwise, but is scheduled to change \
-    in a future release. For the time being, adding hints outside of sections \
-    without specifying an explicit locality attribute is therefore deprecated. It is \
-    recommended to use \"export\" whenever possible. Use the attributes \
+    ~default:CWarnings.AsError
+    (fun () -> strbrk "The default value for hint locality is \
+    currently \"global\" outside sections, but is scheduled to change to \
+    \"export\" in the next release (Coq 8.18). In Coq 8.17, not providing \
+    an explicit locality outside sections triggers a fatal warning, to \
+    ensure that hint localities are made explicit before the upcoming \
+    change in the default value. It is recommended to use \"export\" \
+    whenever possible. Use the attributes \
     #[local], #[global] and #[export] depending on your choice. For example: \
     \"#[export] Hint Unfold foo : bar.\"")
 
@@ -1429,7 +1453,7 @@ let prepare_hint env init (sigma,c) =
       let id = next_ident_away_from default_prepare_hint_ident (fun id -> Id.Set.mem id !vars) in
       vars := Id.Set.add id !vars;
       subst := (evar,mkVar id)::!subst;
-      mkNamedLambda (make_annot id Sorts.Relevant) t (iter (replace_term sigma evar (mkVar id) c)) in
+      mkNamedLambda sigma (make_annot id Sorts.Relevant) t (iter (replace_term sigma evar (mkVar id) c)) in
   let c' = iter c in
     let diff = Univ.ContextSet.diff (Evd.universe_context_set sigma) (Evd.universe_context_set init) in
     (c', diff)
@@ -1548,8 +1572,12 @@ let pr_hint env sigma h = match h.obj with
     str "(*external*) " ++ Pputils.pr_glb_generic env sigma tac
 
 let pr_id_hint env sigma (id, v) =
-  let pr_pat p = str", pattern " ++ pr_lconstr_pattern_env env sigma p in
-  (pr_hint env sigma v.code ++ str"(level " ++ int v.pri ++ pr_opt_no_spc pr_pat v.pat
+  let pr_pat p = match p.pat with
+  | None -> mt ()
+  | Some (ConstrPattern p) -> str", pattern " ++ pr_lconstr_pattern_env env sigma p
+  | Some DefaultPattern -> str", pattern " ++ pr_leconstr_env env sigma (get_default_pattern v.code.obj)
+  in
+  (pr_hint env sigma v.code ++ str"(level " ++ int v.pri ++ pr_pat v
    ++ str", id " ++ int id ++ str ")" ++ spc ())
 
 let pr_hint_list env sigma hintlist =
@@ -1747,7 +1775,10 @@ module FullHint =
 struct
   type t = full_hint
   let priority (h : t) = h.pri
-  let pattern (h : t) = h.pat
+  let pattern (h : t) = match h.pat with
+  | None -> None
+  | Some (ConstrPattern p) -> Some p
+  | Some DefaultPattern -> None (* does not matter, only used by the deprecated Filtered typeclass option *)
   let database (h : t) = h.db
   let run (h : t) k = run_hint h.code k
   let print env sigma (h : t) = pr_hint env sigma h.code
@@ -1764,10 +1795,10 @@ end
 let connect_hint_clenv h gl =
   let { hint_uctx = ctx; hint_clnv = clenv } = h in
   (* [clenv] has been generated by a hint-making function, so the only relevant
-     data in its evarmap is the set of metas. The [evar_reset_evd] function
+     data in its evarmap is the set of metas. The function
      below just replaces the metas of sigma by those coming from the clenv. *)
   let sigma = Tacmach.project gl in
-  let evd = Evd.evars_reset_evd ~with_conv_pbs:true ~with_univs:false sigma clenv.evd in
+  let evd = Evd.meta_merge (Evd.meta_list clenv.evd) (Evd.clear_metas sigma) in
   (* Still, we need to update the universes *)
   match h.hint_uctx with
   | Some ctx ->
