@@ -45,7 +45,7 @@ let interp_fields_evars env sigma ~ninds ~nparams impls_env nots l =
     List.fold_left2
       (fun (env, sigma, uimpls, params, impls_env) no d ->
          let sigma, (i, b, t), impl = match d with
-           | Vernacexpr.AssumExpr({CAst.loc;v=id},bl,t) ->
+           | Vernacexpr.AssumExpr({CAst.v=id},bl,t) ->
              (* Temporary compatibility with the type-classes heuristics *)
              (* which are applied after the interpretation of bl and *)
              (* before the one of t otherwise (see #13166) *)
@@ -53,7 +53,7 @@ let interp_fields_evars env sigma ~ninds ~nparams impls_env nots l =
              let sigma, t, impl =
                ComAssumption.interp_assumption ~program_mode:false env sigma impls_env [] t in
              sigma, (id, None, t), impl
-           | Vernacexpr.DefExpr({CAst.loc;v=id},bl,b,t) ->
+           | Vernacexpr.DefExpr({CAst.v=id},bl,b,t) ->
              let sigma, (b, t), impl =
                ComDefinition.interp_definition ~program_mode:false env sigma impls_env bl None b t in
              let t = match t with Some t -> t | None -> Retyping.get_type_of env sigma b in
@@ -97,10 +97,10 @@ let error_parameters_must_be_named bk {CAst.loc; v=name} =
 let check_parameters_must_be_named = function
   | CLocalDef (b, _, _) ->
     error_parameters_must_be_named default_binder_kind b
-  | CLocalAssum (ls, bk, ce) ->
+  | CLocalAssum (ls, bk, _ce) ->
     List.iter (error_parameters_must_be_named bk) ls
   | CLocalPattern {CAst.loc} ->
-    Loc.raise ?loc (Stream.Error "pattern with quote not allowed in record parameters")
+    Loc.raise ?loc (Gramlib.Stream.Error "pattern with quote not allowed in record parameters")
 
 (** [DataI.t] contains the information used in record interpretation,
    it is a strict subset of [Ast.t] thus this should be
@@ -129,8 +129,11 @@ end
 
 module Data = struct
   type projection_flags = {
-    pf_subclass: bool;
+    pf_coercion: bool;
     pf_reversible: bool;
+    pf_instance: bool;
+    pf_priority: int option;
+    pf_locality: Goptions.option_locality;
     pf_canonical: bool;
   }
   type raw_data = DataR.t
@@ -138,17 +141,17 @@ module Data = struct
   { id : Id.t
   ; idbuild : Id.t
   ; is_coercion : bool
-  ; coers : projection_flags list
+  ; proj_flags : projection_flags list
   ; rdata : raw_data
   ; inhabitant_id : Id.t
   }
 end
 
-let build_type_telescope newps env0 (sigma, template) { DataI.arity; _ } = match arity with
+let build_type_telescope newps env0 sigma { DataI.arity; _ } = match arity with
   | None ->
     let uvarkind = Evd.univ_flexible_alg in
     let sigma, s = Evd.new_sort_variable uvarkind sigma in
-    (sigma, template), (EConstr.mkSort s, s)
+    sigma, (EConstr.mkSort s, s)
   | Some t ->
     let env = EConstr.push_rel_context newps env0 in
     let poly =
@@ -164,15 +167,14 @@ let build_type_telescope newps env0 (sigma, template) { DataI.arity; _ } = match
           match Evd.is_sort_variable sigma s' with
           | Some l ->
             let sigma = Evd.make_flexible_variable sigma ~algebraic:true l in
-            (sigma, template), (s, s')
+            sigma, (s, s')
           | None ->
-            (sigma, false), (s, s')
-        else (sigma, false), (s, s'))
+            sigma, (s, s')
+        else sigma, (s, s'))
      | _ -> user_err ?loc:(constr_loc t) (str"Sort expected."))
 
 type tc_result =
-  bool
-  * Impargs.manual_implicits
+  Impargs.manual_implicits
   (* Part relative to closing the definitions *)
   * UState.named_universes_entry
   * Entries.variance_entry
@@ -190,10 +192,10 @@ let typecheck_params_and_fields def poly udecl ps (records : DataI.t list) : tc_
   let env0 = if not poly && is_template then Environ.set_universes_lbound env0 UGraph.Bound.Prop else env0 in
   let sigma, decl, variances = Constrintern.interp_cumul_univ_decl_opt env0 udecl in
   let () = List.iter check_parameters_must_be_named ps in
-  let sigma, (impls_env, ((env1,newps), imps)) =
+  let sigma, (impls_env, ((_env1,newps), imps)) =
     Constrintern.interp_context_evars ~program_mode:false env0 sigma ps in
-  let (sigma, template), typs =
-    List.fold_left_map (build_type_telescope newps env0) (sigma, true) records in
+  let sigma, typs =
+    List.fold_left_map (build_type_telescope newps env0) sigma records in
   let arities = List.map (fun (typ, _) -> EConstr.it_mkProd_or_LetIn typ newps) typs in
   let relevances = List.map (fun (_,s) -> Sorts.relevance_of_sort s) typs in
   let fold accu { DataI.name; _ } arity r =
@@ -206,17 +208,17 @@ let typecheck_params_and_fields def poly udecl ps (records : DataI.t list) : tc_
   in
   let ninds = List.length arities in
   let nparams = List.length newps in
-  let fold sigma { DataI.nots; fs; _ } arity =
+  let fold sigma { DataI.nots; fs; _ } =
     interp_fields_evars env_ar sigma ~ninds ~nparams impls_env nots fs
   in
-  let (sigma, data) = List.fold_left2_map fold sigma records arities in
+  let (sigma, data) = List.fold_left_map fold sigma records in
   let sigma =
     Pretyping.solve_remaining_evars Pretyping.all_and_fail_flags env_ar sigma in
   let fold sigma (typ, sort) (_, newfs) =
     let univ = ComInductive.Internal.compute_constructor_level env_ar sigma newfs in
     let univ = if Sorts.is_sprop sort then univ else if Sorts.is_sprop univ then Sorts.prop else univ in
       if not def && is_impredicative_sort env0 sort then
-        sigma, (univ, typ)
+        sigma, (sort, typ)
       else
         let sigma = Evd.set_leq_sort env_ar sigma univ sort in
         if Sorts.is_small univ &&
@@ -241,7 +243,7 @@ let typecheck_params_and_fields def poly udecl ps (records : DataI.t list) : tc_
   let univs = Evd.check_univ_decl ~poly sigma decl in
   let ce t = Pretyping.check_evars env0 sigma (EConstr.of_constr t) in
   let () = List.iter (iter_constr ce) (List.rev newps) in
-  template, imps, univs, variances, newps, ans
+  imps, univs, variances, newps, ans
 
 type record_error =
   | MissingProj of Id.t * Id.t list
@@ -252,9 +254,9 @@ let warn_cannot_define_projection =
          (fun msg -> hov 0 msg)
 
 (* If a projection is not definable, we throw an error if the user
-asked it to be a coercion. Otherwise, we just print an info
+asked it to be a coercion or instance. Otherwise, we just print an info
 message. The user might still want to name the field of the record. *)
-let warning_or_error ~info coe indsp err =
+let warning_or_error ~info flags indsp err =
   let st = match err with
     | MissingProj (fi,projs) ->
         let s,have = if List.length projs > 1 then "s","were" else "","was" in
@@ -262,14 +264,14 @@ let warning_or_error ~info coe indsp err =
            strbrk" cannot be defined because the projection" ++ str s ++ spc () ++
            prlist_with_sep pr_comma Id.print projs ++ spc () ++ str have ++
            strbrk " not defined.")
-    | BadTypedProj (fi,ctx,te) ->
+    | BadTypedProj (fi,_ctx,te) ->
         match te with
-          | ElimArity (_,_,_,Some (_,_,_,NonInformativeToInformative)) ->
+          | ElimArity (_, _, Some (_, _, (InType | InSet), InProp)) ->
               (Id.print fi ++
                 strbrk" cannot be defined because it is informative and " ++
                 Printer.pr_inductive (Global.env()) indsp ++
                 strbrk " is not.")
-          | ElimArity (_,_,_,Some (_,_,_,StrongEliminationOnNonSmallType)) ->
+          | ElimArity (_, _, Some (_, _, InType, InSet)) ->
               (Id.print fi ++
                 strbrk" cannot be defined because it is large and " ++
                 Printer.pr_inductive (Global.env()) indsp ++
@@ -277,7 +279,7 @@ let warning_or_error ~info coe indsp err =
           | _ ->
               (Id.print fi ++ strbrk " cannot be defined because it is not typable.")
   in
-  if coe then user_err ~info st;
+  if flags.Data.pf_coercion || flags.Data.pf_instance then user_err ~info st;
   warn_cannot_define_projection (hov 0 st)
 
 type field_status =
@@ -324,6 +326,28 @@ let instantiate_possibly_recursive_type ind u ntypes paramdecls fields =
 
 (* We build projections *)
 
+(** Declare projection [ref] over [from] a coercion
+   or a typeclass instance according to [flags]. *)
+(* remove the last argument (it will become alway true) after deprecation phase
+   (started in 8.17, c.f. https://github.com/coq/coq/pull/16230) *)
+let declare_proj_coercion_instance ~flags ref from ~poly ~with_coercion =
+  if with_coercion && flags.Data.pf_coercion then begin
+    let cl = ComCoercion.class_of_global from in
+    let local = flags.Data.pf_locality = Goptions.OptLocal in
+    ComCoercion.try_add_new_coercion_with_source ref ~local ~poly ~nonuniform:false ~reversible:flags.Data.pf_reversible ~source:cl
+  end;
+  if flags.Data.pf_instance then begin
+    let env = Global.env () in
+    let sigma = Evd.from_env env in
+    let info = Typeclasses.{ hint_priority = flags.Data.pf_priority; hint_pattern = None } in
+    let local =
+      match flags.Data.pf_locality with
+      | Goptions.OptLocal -> Hints.Local
+      | Goptions.(OptDefault | OptExport) -> Hints.Export
+      | Goptions.OptGlobal -> Hints.SuperGlobal in
+    Classes.declare_instance ~warn:true env sigma (Some info) local ref
+  end
+
 (* TODO: refactor the declaration part here; this requires some
    surgery as Evarutil.finalize is called too early in the path *)
 (** This builds and _declares_ a named projection, the code looks
@@ -375,15 +399,12 @@ let build_named_proj ~primitive ~flags ~poly ~univs ~uinstance ~kind env paramde
     | None ->
       let proj_args = (*Rel 1 refers to "x"*) paramargs@[mkRel 1] in
       match decl with
-      | LocalDef (_,ci,_) when primitive -> body
+      | LocalDef _ when primitive -> body
       | _ -> applist (mkConstU (kn,uinstance),proj_args)
   in
   let refi = GlobRef.ConstRef kn in
   Impargs.maybe_declare_manual_implicits false refi impls;
-  if flags.Data.pf_subclass then begin
-    let cl = ComCoercion.class_of_global (GlobRef.IndRef indsp) in
-    ComCoercion.try_add_new_coercion_with_source refi ~local:false ~poly ~nonuniform:false ~reversible:flags.Data.pf_reversible ~source:cl
-  end;
+  declare_proj_coercion_instance ~flags refi (GlobRef.IndRef indsp) ~poly ~with_coercion:true;
   let i = if is_local_assum decl then i+1 else i in
   (Some kn, i, Projection term::subst)
 
@@ -403,7 +424,7 @@ let build_proj env mib indsp primitive x rp lifted_fields ~poly paramdecls param
             subst nfi ti i indsp mib lifted_fields x rp
       with NotDefinable why as exn ->
         let _, info = Exninfo.capture exn in
-        warning_or_error ~info flags.Data.pf_subclass indsp why;
+        warning_or_error ~info flags indsp why;
         (None,i,NoProjection fi::subst)
   in
   (nfi - 1, i,
@@ -444,39 +465,7 @@ let declare_projections indsp univs ?(kind=Decls.StructureComponent) inhabitant_
 
 open Typeclasses
 
-let check_template ~template ~poly ~univs ~params { Data.id; rdata = { DataR.min_univ; fields; _ }; _ } =
-  let template_candidate () =
-    (* we use some dummy values for the arities in the rel_context
-       as univs_of_constr doesn't care about localassums and
-       getting the real values is too annoying *)
-    let add_levels c levels = Univ.Level.Set.union levels (Vars.universes_of_constr c) in
-    let param_levels =
-      List.fold_left (fun levels d -> match d with
-          | LocalAssum _ -> levels
-          | LocalDef (_,b,t) -> add_levels b (add_levels t levels))
-        Univ.Level.Set.empty params
-    in
-    let ctor_levels = List.fold_left
-        (fun univs d ->
-           let univs =
-             RelDecl.fold_constr (fun c univs -> add_levels c univs) d univs
-           in
-           univs)
-        param_levels fields
-    in
-    ComInductive.template_polymorphism_candidate ~ctor_levels univs params
-      (Some min_univ)
-  in
-  match template with
-  | Some template, _ ->
-    (* templateness explicitly requested *)
-    if poly && template then user_err Pp.(strbrk "Attributes \"template\" and \"polymorphism\" are not compatible.");
-    template
-  | None, template ->
-    (* auto detect template *)
-    ComInductive.should_auto_template id (template && template_candidate ())
-
-let load_structure i structure = Structure.register structure
+let load_structure _ structure = Structure.register structure
 
 let cache_structure o = load_structure 1 o
 
@@ -492,7 +481,7 @@ let inStruc : Structure.t -> Libobject.obj =
     cache_function = cache_structure;
     load_function = load_structure;
     subst_function = subst_structure;
-    classify_function = (fun x -> Substitute);
+    classify_function = (fun _ -> Substitute);
     discharge_function = discharge_structure;
     rebuild_function = rebuild_structure; }
 
@@ -567,7 +556,7 @@ module Ast = struct
     ; default_inhabitant_id : Id.t option
     }
 
-  let to_datai { name; is_coercion; cfs; idbuild; sort } =
+  let to_datai { name; cfs; sort; _ } =
     let fs = List.map fst cfs in
     { DataI.name = name.CAst.v
     ; arity = sort
@@ -589,9 +578,16 @@ let check_unique_names records =
   | [] -> ()
   | id :: _ -> user_err (str "Two objects have the same name" ++ spc () ++ quote (Id.print id) ++ str ".")
 
+type kind_class = NotClass | RecordClass | DefClass
+
+let kind_class =
+  let open Vernacexpr in
+  function Class true -> DefClass | Class false -> RecordClass
+  | Inductive_kw | CoInductive | Variant | Record | Structure -> NotClass
+
 let check_priorities kind records =
   let open Vernacexpr in
-  let isnot_class = match kind with Class false -> false | _ -> true in
+  let isnot_class = kind_class kind <> RecordClass in
   let has_priority { Ast.cfs; _ } =
     List.exists (fun (_, { rf_priority }) -> not (Option.is_empty rf_priority)) cfs
   in
@@ -614,12 +610,72 @@ let extract_record_data records =
   in
   ps, data
 
-let pre_process_structure udecl kind ~template ~cumulative ~poly ~primitive_proj finite (records : Ast.t list) =
+let implicits_of_context ctx =
+  List.map (fun name -> CAst.make (Some (name,true)))
+    (List.rev (Anonymous :: (List.filter_map (function
+         | LocalDef _ -> None
+         | LocalAssum _ as d -> Some (RelDecl.get_name d))
+         ctx)))
+
+(* deprecated in 8.16, to be removed at the end of the deprecation phase
+   (c.f., https://github.com/coq/coq/pull/15802 ) *)
+let warn_future_coercion_class_constructor =
+  CWarnings.create ~name:"future-coercion-class-constructor" ~category:"records"
+    ~default:CWarnings.AsError
+    Pp.(fun () -> str "'Class >' currently does nothing. Use 'Class' instead.")
+
+(* deprecated in 8.17, to be removed at the end of the deprecation phase
+   (c.f., https://github.com/coq/coq/pull/16230 ) *)
+let warn_future_coercion_class_field =
+  CWarnings.create ~name:"future-coercion-class-field" ~category:"records" Pp.(fun definitional ->
+    strbrk "A coercion will be introduced instead of an instance in future versions when using ':>' in 'Class' declarations. "
+    ++ strbrk "Replace ':>' with '::' (or use '#[global] Existing Instance field.' for compatibility with Coq < 8.17). Beware that the default locality for '::' is #[export], as opposed to #[global] for ':>' currently."
+    ++ strbrk (if definitional then "" else " Add an explicit #[global] attribute to the field if you need to keep the current behavior. For example: \"Class foo := { #[global] field :: bar }.\""))
+
+let check_proj_flags kind rf =
   let open Vernacexpr in
+  let pf_coercion, pf_reversible =
+    match rf.rf_coercion with
+    (* replace "kind_class kind = NotClass" with true after deprecation phase *)
+    | AddCoercion -> kind_class kind = NotClass, Option.default true rf.rf_reversible
+    | NoCoercion ->
+       if rf.rf_reversible <> None then
+         Attributes.(unsupported_attributes
+           [CAst.make ("reversible (without :>)",VernacFlagEmpty)]);
+       false, false in
+  let pf_instance =
+    match rf.rf_instance with NoInstance -> false | BackInstance -> true
+    | BackInstanceWarning -> kind_class kind <> NotClass in
+  let pf_priority = rf.rf_priority in
+  let pf_locality =
+    begin match rf.rf_coercion, rf.rf_instance with
+    | NoCoercion, NoInstance ->
+       if rf.rf_locality <> Goptions.OptDefault then
+         Attributes.(unsupported_attributes
+           [CAst.make ("locality (without :> or ::)",VernacFlagEmpty)])
+    | AddCoercion, NoInstance ->
+       if rf.rf_locality = Goptions.OptExport then
+         Attributes.(unsupported_attributes
+           [CAst.make ("export (without ::)",VernacFlagEmpty)])
+    | _ -> ()
+    end; rf.rf_locality in
+  (* remove following let after deprecation phase (started in 8.17,
+     c.f., https://github.com/coq/coq/pull/16230 ) *)
+  let pf_locality =
+    match rf.rf_instance, rf.rf_locality with
+    | BackInstanceWarning, Goptions.OptDefault -> Goptions.OptGlobal
+    | _ -> pf_locality in
+  let pf_canonical = rf.rf_canonical in
+  Data.{ pf_coercion; pf_reversible; pf_instance; pf_priority; pf_locality; pf_canonical }
+
+(* remove the definitional argument at the end of the deprecation phase
+   (started in 8.17)
+   (c.f., https://github.com/coq/coq/pull/16230 ) *)
+let pre_process_structure ?(definitional=false) udecl kind ~poly (records : Ast.t list) =
   let () = check_unique_names records in
   let () = check_priorities kind records in
   let ps, data = extract_record_data records in
-  let auto_template, impargs, univs, variances, params, data =
+  let impargs, univs, variances, params, data =
     (* In theory we should be able to use
        [Notation.with_notation_protection], due to the call to
        Metasyntax.set_notation_for_interpretation, however something
@@ -628,10 +684,32 @@ let pre_process_structure udecl kind ~template ~cumulative ~poly ~primitive_proj
     Vernacstate.System.protect (fun () ->
         typecheck_params_and_fields (kind = Class true) poly udecl ps data) ()
   in
-  let template = template, auto_template in
-  template, impargs, params, univs, variances, data
+  let adjust_impls impls = match kind_class kind with
+    | NotClass -> impargs @ [CAst.make None] @ impls
+    | _ -> implicits_of_context params @ impls in
+  let data = List.map (fun ({ DataR.implfs; _ } as d) -> { d with DataR.implfs = List.map adjust_impls implfs }) data in
+  let map rdata { Ast.name; is_coercion; cfs; idbuild; default_inhabitant_id; _ } =
+    let proj_flags = List.map (fun (_, rf) -> check_proj_flags kind rf) cfs in
+    let inhabitant_id =
+      match default_inhabitant_id, kind_class kind with
+      | None, NotClass -> data_name name.CAst.v rdata
+      | None, _ -> Namegen.next_ident_away name.CAst.v (Termops.vars_of_env (Global.env()))
+      | Some n, _ -> n
+    in
+    let is_coercion = match is_coercion with AddCoercion -> true | NoCoercion -> false in
+    if kind_class kind <> NotClass then begin
+      if is_coercion then warn_future_coercion_class_constructor ();
+      if List.exists (function (_, Vernacexpr.{ rf_instance = BackInstanceWarning; _ }) -> true | _ -> false) cfs then
+        warn_future_coercion_class_field definitional
+    end;
+    { Data.id = name.CAst.v; idbuild; rdata; is_coercion; proj_flags; inhabitant_id }
+  in
+  let data = List.map2 map data records in
+  let projections_kind =
+    Decls.(match kind_class kind with NotClass -> StructureComponent | _ -> Method) in
+  impargs, params, univs, variances, projections_kind, data
 
-let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind records data =
+let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind data =
   let nparams = List.length params in
   let (univs, ubinders) = univs in
   let poly, projunivs =
@@ -640,7 +718,7 @@ let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj i
     | UState.Polymorphic_entry uctx -> true, Entries.Polymorphic_entry uctx
   in
   let ntypes = List.length data in
-  let mk_block i { Data.id; idbuild; rdata = { DataR.min_univ; arity; fields; _ }; _ } =
+  let mk_block i { Data.id; idbuild; rdata = { DataR.arity; fields; _ }; _ } =
     let nfields = List.length fields in
     let args = Context.Rel.instance_list mkRel nfields params in
     let ind = applist (mkRel (ntypes - i + nparams + nfields), args) in
@@ -651,24 +729,31 @@ let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj i
       mind_entry_lc = [type_constructor] }
   in
   let blocks = List.mapi mk_block data in
-  let template = match data, fst template with
-    | [data], _ -> check_template ~template ~univs ~poly ~params data
-    | _, Some true -> user_err Pp.(str "Template-polymorphism not allowed with mutual records.")
-    | _ -> false
+  let ind_univs, global_univ_decls = match blocks, data with
+  | [entry], [data] ->
+    let concl = Some data.Data.rdata.DataR.min_univ in
+    let env_ar_params = Environ.push_rel_context params (Global.env ()) in
+    ComInductive.compute_template_inductive ~user_template:template
+      ~env_ar_params ~ctx_params:params ~univ_entry:univs entry concl
+  | _ ->
+    begin match template with
+    | Some true -> user_err Pp.(str "Template-polymorphism not allowed with mutual records.")
+    | Some false | None ->
+      match univs with
+      | UState.Polymorphic_entry uctx -> Polymorphic_ind_entry uctx, Univ.ContextSet.empty
+      | UState.Monomorphic_entry uctx -> Monomorphic_ind_entry, uctx
+    end
   in
   let primitive =
     primitive_proj  &&
     List.for_all (fun { Data.rdata = { DataR.fields; _ }; _ } -> List.exists is_local_assum fields) data
   in
-  let globnames, univs, global_univ_decls = match univs with
-  | UState.Monomorphic_entry ctx ->
-    if template then
-      (univs, ubinders), Template_ind_entry ctx, None
-    else
-      (univs, ubinders), Monomorphic_ind_entry, Some ctx
-  | UState.Polymorphic_entry ctx ->
-    (univs, UnivNames.empty_binders), Polymorphic_ind_entry ctx, None
+  let globnames, global_univ_decls = match ind_univs with
+  | Monomorphic_ind_entry -> (univs, ubinders), Some global_univ_decls
+  | Template_ind_entry _ -> (univs, ubinders), Some global_univ_decls
+  | Polymorphic_ind_entry _ -> (univs, UnivNames.empty_binders), None
   in
+  let univs = ind_univs in
   let variance = ComInductive.variance_of_entry ~cumulative ~variances univs in
   let mie =
     { mind_entry_params = params;
@@ -686,57 +771,45 @@ let interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj i
 
 
 let interp_structure udecl kind ~template ~cumulative ~poly ~primitive_proj finite records =
-  let open Vernacexpr in
-  let template, impargs, params, univs, variances, data =
-    pre_process_structure udecl kind ~template ~cumulative ~poly ~primitive_proj finite records in
-  let adjust_impls impls = impargs @ [CAst.make None] @ impls in
-  let data = List.map (fun ({ DataR.implfs; _ } as d) -> { d with DataR.implfs = List.map adjust_impls implfs }) data in
-  (* let map (min_univ, arity, fieldimpls, fields) { Ast.name; is_coercion; cfs; idbuild; _ } = *)
-  let map rdata { Ast.name; is_coercion; cfs; idbuild; default_inhabitant_id; _ } =
-    let coers = List.map (fun (_, { rf_subclass ; rf_reverse ; rf_canonical }) ->
-      let pf_subclass, pf_reversible =
-        match rf_subclass with
-        | Vernacexpr.BackInstance -> true, Option.default true rf_reverse
-        | Vernacexpr.NoInstance ->
-          if rf_reverse <> None then
-            Attributes.(unsupported_attributes
-              [CAst.make ("reversible (without :>)",VernacFlagEmpty)]);
-          false, false in
-      { Data.pf_subclass; pf_reversible; pf_canonical = rf_canonical })
-      cfs
-    in
-    let inhabitant_id =
-      match default_inhabitant_id with
-      | None -> data_name name.CAst.v rdata
-      | Some n -> n
-    in
-    { Data.id = name.CAst.v; idbuild; rdata; is_coercion; coers; inhabitant_id }
-  in
-  let data = List.map2 map data records in
-  interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind:Decls.StructureComponent records data
+  let impargs, params, univs, variances, projections_kind, data =
+    pre_process_structure udecl kind ~poly records in
+  interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind data
 
 let declare_structure { Record_decl.mie; primitive_proj; impls; globnames; global_univ_decls; projunivs; ubinders; projections_kind; poly; records } =
   Option.iter (DeclareUctx.declare_universe_context ~poly:false) global_univ_decls;
   let kn = DeclareInd.declare_mutual_inductive_with_eliminations mie globnames impls
       ~primitive_expected:primitive_proj
   in
-  let map i { Data.is_coercion; coers; rdata = { DataR.implfs; fields; _}; inhabitant_id; id; _ } =
+  let map i { Data.is_coercion; proj_flags; rdata = { DataR.implfs; fields; _}; inhabitant_id; _ } =
     let rsp = (kn, i) in (* This is ind path of idstruc *)
     let cstr = (rsp, 1) in
-    let projections = declare_projections rsp (projunivs,ubinders) ~kind:projections_kind inhabitant_id coers implfs fields in
+    let projections = declare_projections rsp (projunivs,ubinders) ~kind:projections_kind inhabitant_id proj_flags implfs fields in
     let build = GlobRef.ConstructRef cstr in
     let () = if is_coercion then ComCoercion.try_add_new_coercion build ~local:false ~poly ~nonuniform:false ~reversible:true in
     let struc = Structure.make (Global.env ()) rsp projections in
     let () = declare_structure_entry struc in
-    rsp
+    GlobRef.IndRef rsp
   in
-  List.mapi map records
+  List.mapi map records, []
 
-let implicits_of_context ctx =
-  List.map (fun name -> CAst.make (Some (name,true)))
-    (List.rev (Anonymous :: (List.map RelDecl.get_name ctx)))
+let get_class_params : Data.t list -> Data.t = function
+  | [data] -> data
+  | _ ->
+    CErrors.user_err (str "Mutual definitional classes are not supported.")
 
-let build_class_constant ~univs ~rdata ~primitive_proj field implfs params paramimpls coers binder id proj_name =
+(* declare definitional class (typeclasses that are not record) *)
+(* [data] is a list with a single [Data.t] with a single field (in [Data.rdata])
+   and [Data.is_coercion] must be [NoCoercion] *)
+let declare_class_constant ~univs paramimpls params data =
+  let {Data.id; rdata; is_coercion; proj_flags; inhabitant_id} = get_class_params data in
+  assert (not is_coercion);  (* should be ensured by caller *)
+  let implfs = rdata.DataR.implfs in
+  let field, binder, proj_name, proj_flags = match rdata.DataR.fields, proj_flags with
+    | [ LocalAssum ({binder_name=Name proj_name} as binder, field)
+      | LocalDef ({binder_name=Name proj_name} as binder, _, field) ], [proj_flags] ->
+      let binder = {binder with binder_name=Name inhabitant_id} in
+      field, binder, proj_name, proj_flags
+    | _ -> assert false in  (* should be ensured by caller *)
   let class_body = it_mkLambda_or_LetIn field params in
   let class_type = it_mkProd_or_LetIn rdata.DataR.arity params in
   let class_entry =
@@ -751,8 +824,7 @@ let build_class_constant ~univs ~rdata ~primitive_proj field implfs params param
       Univ.UContext.instance uctx, univs
   in
   let cstu = (cst, inst) in
-  let inst_type = appvectc (mkConstU cstu)
-      (Termops.rel_vect 0 (List.length params)) in
+  let inst_type = appvectc (mkConstU cstu) (Context.Rel.instance mkRel 0 params) in
   let proj_type =
     it_mkProd_or_LetIn (mkProd(binder, inst_type, lift 1 field)) params in
   let proj_body =
@@ -766,51 +838,24 @@ let build_class_constant ~univs ~rdata ~primitive_proj field implfs params param
   Impargs.declare_manual_implicits false (GlobRef.ConstRef proj_cst) (List.hd implfs);
   Classes.set_typeclass_transparency ~locality:Hints.SuperGlobal
     [Tacred.EvalConstRef cst] false;
-  let sub = fst (List.hd coers) in
+  let () =
+    let csb = Global.lookup_constant cst in
+    let poly = Declareops.constant_is_polymorphic csb in
+    declare_proj_coercion_instance ~flags:proj_flags (GlobRef.ConstRef proj_cst) cref ~poly ~with_coercion:false in
   let m = {
     meth_name = Name proj_name;
-    meth_info = sub;
+    meth_info = None;
     meth_const = Some proj_cst;
   } in
-  [cref, [m]]
-
-let build_record_constant ~rdata ~univs ~variances ~cumulative ~template ~primitive_proj
-    fields params paramimpls is_coercion coers id idbuild inhabitant_id =
-  let record_data =
-    { Data.id
-    ; idbuild
-    ; is_coercion = false
-    (* to be replaced by the following line after deprecation phase
-       (started in 8.16, c.f., https://github.com/coq/coq/pull/15802 ) *)
-    (* ; is_coercion *)
-    ; coers = List.map (fun _ -> { Data.pf_subclass = false ; pf_reversible = false ; pf_canonical = true }) coers
-    (* to be replaced by the following line after deprecation phase
-       (started in 8.16, c.f., https://github.com/coq/coq/pull/15802 ) *)
-    (* ; coers = List.map (fun (c, rev) -> { pf_subclass = c <> None ; pf_reversible = rev ; pf_canonical = true }) coers *)
-    ; rdata
-    ; inhabitant_id
-    } in
-  let structure =
-    interp_structure_core  ~cumulative Declarations.BiFinite ~univs ~variances ~primitive_proj paramimpls
-      params template ~projections_kind:Decls.Method rdata [record_data] in
-  let inds = declare_structure structure in
-  let map ind =
-    let map decl (b, _) y = {
-      meth_name = RelDecl.get_name decl;
-      meth_info = b;
-      meth_const = y;
-    } in
-    let l = List.map3 map (List.rev fields) coers (Structure.find_projections ind) in
-    GlobRef.IndRef ind, l
-  in
-  List.map map inds
+  [cref], [m]
 
 (** [declare_class] will prepare and declare a [Class]. This is done in
    2 steps:
 
   1. two markedly different paths are followed depending on whether the
-   class declaration refers to a constant "definitional classes" or to
-   a record, that is to say:
+   class declaration refers to a constant "definitional classes"
+   (with [declare_class_constant]) or to a record (with [declare_structure]),
+   that is to say:
 
       Class foo := bar : T.
 
@@ -824,29 +869,26 @@ let build_record_constant ~rdata ~univs ~variances ~cumulative ~template ~primit
 
       Class foo := { ... }.
 
-  2. declare the class, using the information from 1. in the form of [Classes.typeclass]
+  2. now, declare the class, using the information ([inds] and [def]) from 1.
+   in the form of [Classes.typeclass]
 
   *)
-let declare_class def ~cumulative ~univs ~variances ~primitive_proj id idbuild inhabitant_id paramimpls params
-    rdata template ?(kind=Decls.StructureComponent) is_coercion coers =
-  let implfs =
-    (* Make the class implicit in the projections, and the params if applicable. *)
-    let impls = implicits_of_context params in
-      List.map (fun x -> impls @ x) rdata.DataR.implfs
-  in
-  let rdata = { rdata with DataR.implfs } in
+let declare_class ~univs params inds def data =
+  let { Data.rdata } = get_class_params data in
   let fields = rdata.DataR.fields in
-  let data =
-    match fields with
-    | [ LocalAssum ({binder_name=Name proj_name} as binder, field)
-      | LocalDef ({binder_name=Name proj_name} as binder, _, field) ] when def ->
-      assert (not is_coercion);  (* should be ensured by caller *)
-      let binder = {binder with binder_name=Name inhabitant_id} in
-      build_class_constant ~rdata ~univs ~primitive_proj field implfs params paramimpls coers binder id proj_name
-    | _ ->
-      build_record_constant ~rdata ~univs ~variances ~cumulative ~template ~primitive_proj
-        fields params paramimpls is_coercion coers id idbuild inhabitant_id
+  let map ind =
+    let map decl y = {
+      meth_name = RelDecl.get_name decl;
+      meth_info = None;
+      meth_const = y;
+    } in
+    let l = match ind with
+      | GlobRef.IndRef ind ->
+         List.map2 map (List.rev fields) (Structure.find_projections ind)
+      | _ -> def in
+    ind, l
   in
+  let data = List.map map inds in
   let univs, params, fields =
     match fst univs with
     | UState.Polymorphic_entry uctx ->
@@ -869,13 +911,12 @@ let declare_class def ~cumulative ~univs ~variances ~primitive_proj id idbuild i
         cl_props = fields;
         cl_projs = projs }
     in
-    let env = Global.env () in
-    let sigma = Evd.from_env env in
-    Classes.add_class env sigma k; impl
+    Classes.add_class k
   in
-  List.map map data
+  List.iter map data
 
-let add_constant_class env sigma cst =
+let add_constant_class cst =
+  let env = Global.env () in
   let ty, univs = Typeops.type_of_global_in_context env (GlobRef.ConstRef cst) in
   let r = (Environ.lookup_constant cst env).const_relevance in
   let ctx, _ = decompose_prod_assum ty in
@@ -891,11 +932,12 @@ let add_constant_class env sigma cst =
       cl_unique = typeclasses_unique ()
     }
   in
-  Classes.add_class env sigma tc;
+  Classes.add_class tc;
   Classes.set_typeclass_transparency ~locality:Hints.SuperGlobal
     [Tacred.EvalConstRef cst] false
 
-let add_inductive_class env sigma ind =
+let add_inductive_class ind =
+  let env = Global.env () in
   let mind, oneind = Inductive.lookup_mind_specif env ind in
   let k =
     let ctx = oneind.mind_arity_ctxt in
@@ -911,80 +953,51 @@ let add_inductive_class env sigma ind =
         cl_strict = typeclasses_strict ();
         cl_unique = typeclasses_unique () }
   in
-  Classes.add_class env sigma k
+  Classes.add_class k
 
 let warn_already_existing_class =
   CWarnings.create ~name:"already-existing-class" ~category:"automation" Pp.(fun g ->
       Printer.pr_global g ++ str " is already declared as a typeclass.")
 
 let declare_existing_class g =
-  let env = Global.env () in
-  let sigma = Evd.from_env env in
   if Typeclasses.is_class g then warn_already_existing_class g
   else
     match g with
-    | GlobRef.ConstRef x -> add_constant_class env sigma x
-    | GlobRef.IndRef x -> add_inductive_class env sigma x
+    | GlobRef.ConstRef x -> add_constant_class x
+    | GlobRef.IndRef x -> add_inductive_class x
     | _ -> user_err
              (Pp.str"Unsupported class type, only constants and inductives are allowed.")
 
-open Vernacexpr
-
-(* deprecated in 8.16, to be removed at the end of the deprecation phase
-   (c.f., https://github.com/coq/coq/pull/15802 ) *)
-let warn_future_coercion_class_constructor =
-  CWarnings.create ~name:"future-coercion-class-constructor" ~category:"records"
-    ~default:CWarnings.AsError
-    Pp.(fun () -> str "'Class >' currently does nothing. Use 'Class' instead.")
-
-(* declaring structures, common data to refactor *)
-let class_structure udecl kind def ~template ~cumulative ~poly ~primitive_proj finite records =
-  let template, impargs, params, univs, variances, data =
-    pre_process_structure udecl kind ~template ~cumulative ~poly ~primitive_proj finite records in
-  let { Ast.name; is_coercion; cfs; idbuild; default_inhabitant_id; _ }, rdata = match records, data with
-    | [r], [d] -> r, d
-    | _, _ ->
-      CErrors.user_err (str "Mutual definitional classes are not handled.")
-  in
-  if is_coercion then warn_future_coercion_class_constructor ();
-  let coers = List.map (fun (_, { rf_subclass; rf_reverse; rf_priority }) ->
-      if rf_reverse <> None then
-        Attributes.(unsupported_attributes
-          [CAst.make ("reversible (without coercion)",VernacFlagEmpty)]);
-      match rf_subclass with
-      | Vernacexpr.BackInstance -> Some {hint_priority = rf_priority; hint_pattern = None}, Option.default true rf_reverse
-      | Vernacexpr.NoInstance -> None, false)
-      cfs
-  in
-  let inhabitant_id =
-    match default_inhabitant_id with
-    | None -> Namegen.next_ident_away name.CAst.v (Termops.vars_of_env (Global.env()))
-    | Some id -> id
-  in
-  declare_class def ~cumulative ~univs ~primitive_proj ~variances name.CAst.v idbuild inhabitant_id
-    impargs params rdata template is_coercion coers
-
-let regular_structure udecl kind ~template ~cumulative ~poly ~primitive_proj finite records =
-  let structure = interp_structure udecl kind ~template ~cumulative ~poly ~primitive_proj finite records in
-  let inds = declare_structure structure in
-  List.map (fun ind -> GlobRef.IndRef ind) inds
-
-(** [fs] corresponds to fields and [ps] to parameters; [coers] is a
+(** [fs] corresponds to fields and [ps] to parameters; [proj_flags] is a
     list telling if the corresponding fields must me declared as coercions
     or subinstances. *)
 
 let definition_structure udecl kind ~template ~cumulative ~poly ~primitive_proj
     finite (records : Ast.t list) : GlobRef.t list =
-  match kind with
-  | Class def ->
-    class_structure udecl kind def ~template ~cumulative ~poly ~primitive_proj finite records
-  | Inductive_kw | CoInductive | Variant | Record | Structure ->
-    regular_structure udecl kind ~template ~cumulative ~poly ~primitive_proj finite records
+  let impargs, params, univs, variances, projections_kind, data =
+    let definitional = kind_class kind = DefClass in
+    pre_process_structure ~definitional udecl kind ~poly records
+  in
+  let inds, def = match kind_class kind with
+    | DefClass -> declare_class_constant ~univs impargs params data
+    | RecordClass | NotClass ->
+      (* remove the following block after deprecation phase
+         (started in 8.16, c.f., https://github.com/coq/coq/pull/15802 ) *)
+      let data = if kind_class kind = NotClass then data else
+          List.map (fun d -> { d with Data.is_coercion = false }) data in
+      let structure = interp_structure_core ~cumulative finite ~univs ~variances ~primitive_proj impargs params template ~projections_kind data in
+      declare_structure structure
+  in
+  if kind_class kind <> NotClass then declare_class ~univs params inds def data;
+  inds
 
 module Internal = struct
   type nonrec projection_flags = Data.projection_flags = {
-    pf_subclass: bool;
+    pf_coercion: bool;
     pf_reversible: bool;
+    pf_instance: bool;
+    pf_priority: int option;
+    pf_locality: Goptions.option_locality;
     pf_canonical: bool;
   }
   let declare_projections = declare_projections
